@@ -13,27 +13,38 @@ use serde::Deserialize;
 /// with expansion of certain variables.
 ///
 /// The supported variables are:
-///   * `APP_CACHE`: an arti-specific cache directory.
-///   * `APP_CONFIG`: an arti-specific configuration directory.
-///   * `APP_SHARED_DATA`: an arti-specific directory in the user's "shared
+///   * `ARTI_CACHE`: an arti-specific cache directory.
+///   * `ARTI_CONFIG`: an arti-specific configuration directory.
+///   * `ARTI_SHARED_DATA`: an arti-specific directory in the user's "shared
 ///     data" space.
-///   * `APP_LOCAL_DATA`: an arti-specific directory in the user's "local
+///   * `ARTI_LOCAL_DATA`: an arti-specific directory in the user's "local
 ///     data" space.
 ///   * `USER_HOME`: the user's home directory.
 ///
 /// These variables are implemented using the `directories` crate, and
 /// so should use appropriate system-specific overrides under the
-/// hood.
-#[derive(Clone, Debug, Deserialize)]
+/// hood. For more information, see that crate's documentation.
+#[derive(Clone, Debug, Deserialize, Eq, PartialEq)]
 #[serde(transparent)]
-pub struct CfgPath(String);
+pub struct CfgPath(PathInner);
+
+/// Inner implementation of CfgPath
+#[derive(Clone, Debug, Deserialize, Eq, PartialEq)]
+#[serde(untagged)]
+enum PathInner {
+    /// A path that should be expanded from a string using ShellExpand.
+    Shell(String),
+    /// A path that should be used literally, with no expansion.
+    Literal(PathBuf),
+}
 
 /// An error that has occurred while expanding a path.
 #[derive(thiserror::Error, Debug, Clone)]
-pub enum Error {
+#[non_exhaustive]
+pub enum CfgPathError {
     /// The path contained a variable we didn't recognize.
-    #[error("unrecognized variable")]
-    UnknownVar,
+    #[error("unrecognized variable {0}")]
+    UnknownVar(String),
     /// We couldn't construct a ProjectDirs object.
     #[error("can't construct project directories")]
     NoProjectDirs,
@@ -44,38 +55,65 @@ pub enum Error {
     ///
     /// (This is due to a limitation in the shellexpand crate, which should
     /// be fixed in the future.)
-    #[error("can't convert to UTF-8")]
-    BadUtf8,
+    #[error("can't convert value of {0} to UTF-8")]
+    BadUtf8(String),
+    /// We couldn't convert a string to a valid path on the OS.
+    #[error("invalid path string: {0:?}")]
+    InvalidString(String),
 }
 
 impl CfgPath {
     /// Create a new configuration path
     pub fn new(s: String) -> Self {
-        CfgPath(s)
+        CfgPath(PathInner::Shell(s))
     }
 
-    /// Return the path on disk designated by this path.
-    pub fn path(&self) -> Result<PathBuf, shellexpand::LookupError<Error>> {
-        Ok(shellexpand::full_with_context(&self.0, get_home, get_env)?
-            .into_owned()
-            .into())
+    /// Return the path on disk designated by this `CfgPath`.
+    pub fn path(&self) -> Result<PathBuf, CfgPathError> {
+        match &self.0 {
+            PathInner::Shell(s) => expand(s),
+            PathInner::Literal(p) => Ok(p.clone()),
+        }
+    }
+
+    /// Construct a new `CfgPath` from a system path.
+    pub fn from_path<P: AsRef<Path>>(path: P) -> Self {
+        CfgPath(PathInner::Literal(path.as_ref().to_owned()))
     }
 }
 
+/// Helper: expand a directory given as a string.
+#[cfg(feature = "expand-paths")]
+fn expand(s: &str) -> Result<PathBuf, CfgPathError> {
+    Ok(shellexpand::full_with_context(s, get_home, get_env)
+        .map_err(|e| e.cause)?
+        .into_owned()
+        .into())
+}
+
+/// Helper: convert a string to a path without expansion.
+#[cfg(not(feature = "expand-paths"))]
+fn expand(s: &str) -> Result<PathBuf, CfgPathError> {
+    s.try_into()
+        .map_err(|_| CfgPathError::InvalidString(s.to_owned()))
+}
+
 /// Shellexpand helper: return the user's home directory if we can.
+#[cfg(feature = "expand-paths")]
 fn get_home() -> Option<&'static Path> {
     base_dirs().ok().map(BaseDirs::home_dir)
 }
 
 /// Shellexpand helper: Expand a shell variable if we can.
-fn get_env(var: &str) -> Result<Option<&'static str>, Error> {
+#[cfg(feature = "expand-paths")]
+fn get_env(var: &str) -> Result<Option<&'static str>, CfgPathError> {
     let path = match var {
-        "APP_CACHE" => project_dirs()?.cache_dir(),
-        "APP_CONFIG" => project_dirs()?.config_dir(),
-        "APP_SHARED_DATA" => project_dirs()?.data_dir(),
-        "APP_LOCAL_DATA" => project_dirs()?.data_local_dir(),
+        "ARTI_CACHE" => project_dirs()?.cache_dir(),
+        "ARTI_CONFIG" => project_dirs()?.config_dir(),
+        "ARTI_SHARED_DATA" => project_dirs()?.data_dir(),
+        "ARTI_LOCAL_DATA" => project_dirs()?.data_local_dir(),
         "USER_HOME" => base_dirs()?.home_dir(),
-        _ => return Err(Error::UnknownVar),
+        _ => return Err(CfgPathError::UnknownVar(var.to_owned())),
     };
 
     match path.to_str() {
@@ -85,34 +123,39 @@ fn get_env(var: &str) -> Result<Option<&'static str>, Error> {
         // Note that this error is necessary because shellexpand
         // doesn't currently handle OsStr.  In the future, that might
         // change.
-        None => Err(Error::BadUtf8),
+        None => Err(CfgPathError::BadUtf8(var.to_owned())),
     }
 }
 
 impl std::fmt::Display for CfgPath {
     fn fmt(&self, fmt: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
-        self.0.fmt(fmt)
+        match &self.0 {
+            PathInner::Literal(p) => write!(fmt, "{:?} [exactly]", p),
+            PathInner::Shell(s) => s.fmt(fmt),
+        }
     }
 }
 
 /// Return a ProjectDirs object for the Arti project.
-fn project_dirs() -> Result<&'static ProjectDirs, Error> {
+#[cfg(feature = "expand-paths")]
+fn project_dirs() -> Result<&'static ProjectDirs, CfgPathError> {
     /// lazy cell holding the ProjectDirs object.
     static PROJECT_DIRS: Lazy<Option<ProjectDirs>> =
         Lazy::new(|| ProjectDirs::from("org", "torproject", "Arti"));
 
-    PROJECT_DIRS.as_ref().ok_or(Error::NoProjectDirs)
+    PROJECT_DIRS.as_ref().ok_or(CfgPathError::NoProjectDirs)
 }
 
 /// Return a UserDirs object for the current user.
-fn base_dirs() -> Result<&'static BaseDirs, Error> {
+#[cfg(feature = "expand-paths")]
+fn base_dirs() -> Result<&'static BaseDirs, CfgPathError> {
     /// lazy cell holding the BaseDirs object.
     static BASE_DIRS: Lazy<Option<BaseDirs>> = Lazy::new(BaseDirs::new);
 
-    BASE_DIRS.as_ref().ok_or(Error::NoBaseDirs)
+    BASE_DIRS.as_ref().ok_or(CfgPathError::NoBaseDirs)
 }
 
-#[cfg(test)]
+#[cfg(all(test, feature = "expand-paths"))]
 mod test {
     #![allow(clippy::unwrap_used)]
     use super::*;
@@ -143,8 +186,8 @@ mod test {
 
     #[test]
     fn expand_cache() {
-        let p = CfgPath::new("${APP_CACHE}/example".to_string());
-        assert_eq!(p.to_string(), "${APP_CACHE}/example".to_string());
+        let p = CfgPath::new("${ARTI_CACHE}/example".to_string());
+        assert_eq!(p.to_string(), "${ARTI_CACHE}/example".to_string());
 
         let expected = project_dirs().unwrap().cache_dir().join("example");
         assert_eq!(p.path().unwrap().to_str(), expected.to_str());
@@ -152,9 +195,24 @@ mod test {
 
     #[test]
     fn expand_bogus() {
-        let p = CfgPath::new("${APP_WOMBAT}/example".to_string());
-        assert_eq!(p.to_string(), "${APP_WOMBAT}/example".to_string());
+        let p = CfgPath::new("${ARTI_WOMBAT}/example".to_string());
+        assert_eq!(p.to_string(), "${ARTI_WOMBAT}/example".to_string());
 
-        assert!(p.path().is_err());
+        assert!(matches!(p.path(), Err(CfgPathError::UnknownVar(_))));
+        assert_eq!(
+            &p.path().unwrap_err().to_string(),
+            "unrecognized variable ARTI_WOMBAT"
+        );
+    }
+
+    #[test]
+    fn literal() {
+        let p = CfgPath::from_path(PathBuf::from("${ARTI_CACHE}/literally"));
+        // This doesn't get expanded, since we're using a literal path.
+        assert_eq!(
+            p.path().unwrap().to_str().unwrap(),
+            "${ARTI_CACHE}/literally"
+        );
+        assert_eq!(p.to_string(), "\"${ARTI_CACHE}/literally\" [exactly]");
     }
 }
